@@ -21,7 +21,7 @@
  * You are encouraged to send comments, improvements or suggestions to
  * me at ike-scan@nta-monitor.com.
  *
- * $Id: ike-scan.c,v 1.39 2003/01/17 17:29:09 rsh Exp $
+ * $Id: ike-scan.c,v 1.58 2003/05/10 16:09:03 rsh Exp $
  *
  * ike-scan -- The IKE Scanner
  *
@@ -40,105 +40,14 @@
  * backoff to cope with packet loss.
  * 
  * Use ike-scan --help to display information on the usage and options.
+ * See the README file for full details.
  * 
- * The hosts to scan can be specified on the command line or read from an
- * input file using the --file=<fn> option.  The program can cope with
- * large numbers of hosts limited only by the amount of memory needed to store
- * the list of host_entry structures.  Each host_entry structure requires 52
- * bytes on a 32-bit system, so a class B network (65534 hosts) would require
- * about 3.25 MB for the list.
- * 
- * The program limits the rate at which it sends IKE packets to ensure that
- * it does not overload the network connection.  By default the rate limit
- * is one packet every 80ms which equates to a data rate of 36400 bits per
- * second given the default packet size of 364 bytes.  For faster links, the
- * packet rate can be raised by lowering the minimum packet interval using
- * --interval=<n> which sets the minimum packet interval to n ms.
- * 
- * By default, the main mode packets sent contain one proposal which contains
- * 8 transforms.  These 8 transforms represent all possible combinations of:
- * a) Encryption Algorithm: DES-CBC and 3DES-CBC;
- * b) Hash Algorithm: MD5 and SHA; and
- * c) DH Group: 1 (MODP 768) and 2 (MODP 1024).
- * 
- * It is also possible to specify the Authentication Method with --auth
- * (default is 1 - pre-shared key) and the IKE lifetime in seconds with
- * --lifetime (default is 28800 seconds or 8 hours as recommended by RFC 2407).
- * 
- * This default transform set is designed to be acceptable to most IKE
- * implementations - most will accept at least one of the offered transforms.
- * However, it is often necessary to use a different authentication method
- * (pre-shared key is the most common, but is not always supported) and
- * more rarely it may be necessary to reduce the lifetime.
- * 
- * The default transform set results in a packet data length of 336 bytes which
- * when IP and UDP headers are added gives a total packet size of 364 bytes.
- * 
- * It is also possible to specify a single custom transform with
- * --trans=e,h,a,g where e is the Encryption Algorithm, h the Hash Algorithm,
- * a the Authentication Method and g is the DH Group.  These are specified as
- * values; see RFC 2409 Appendix A for details of which values to use.  For
- * example, --trans=2,3,1,5 would specify Enc=IDEA-CBC, Hash=Tiger, DH Group=5
- * (MODP 1536), Auth=1 (pre-shared key).
- * 
- * If a custom transform is specified, then only a single transform is used,
- * and the packet data length is 84 bytes for a total packet length of 112
- * bytes.  Specifying a custom transform also overrides authentication method
- * (either the default of pre-shared key or as specified with --auth).
- * However, it is still possible to change the IKE lifetime of the custom
- * transform with --lifetime.
- * 
- * A custom transform can be useful in the following situations:
- * a) If none of the transforms in the default transform set is acceptable to
- *    the remote IKE implementation;
- * b) If you know that a particular transform will be acceptable, and you want
- *    to minimise bandwidth use or allow faster scanning rates; or
- * c) If you want to determine exactly which transforms a remote IKE
- *    implementation supports for fingerprinting.
- * 
- * For those hosts that respond, ike-scan records the times of the received
- * IKE responses.  The backoff between IKE responses varies between different
- * IKE implementations and can therefore be used as a fingerprint.  The
- * --showbackoff option is used to display the backoff times for each host
- * which responded.  Note that using the --showbackoff option will cause
- * ike-scan to wait for 60 seconds after the last received packet to ensure
- * that it has seen all of the responses.  This 60 second wait can be
- * altered by specifying a different value to the --showbackoff option.
  */
-#ifdef HAVE_CONFIG_H
-#include "config.h"
-#endif
 
-#include <stdio.h>
-#include <string.h>
-#include <stdlib.h>
-#include <unistd.h>
-#include <ctype.h>
-#include <math.h>
-#include <netdb.h>
-#ifdef HAVE_GETOPT_H
-#include <getopt.h>
-#else
-/* Include getopt.h for the sake of getopt_long.
-   We don't need the declaration of getopt, and it could conflict
-   with something from a system header file, so effectively nullify that.  */
-#define getopt getopt_loser
-#include "getopt.h"
-#undef getopt
-#endif
-#include <errno.h>
-#include <syslog.h>
-#ifdef HAVE_LIBGEN_H
-#include <libgen.h>
-#endif
-
-#include "global.h"
-#include "md5.h"
 #include "ike-scan.h"
 
-#define VERSION "ike-scan v1.0"
 #define MAX_PAYLOAD 13	/* Maximum defined payload number */
-static char rcsid[] = "$Id: ike-scan.c,v 1.39 2003/01/17 17:29:09 rsh Exp $";   /* RCS ID for ident(1) */
+static char rcsid[] = "$Id: ike-scan.c,v 1.58 2003/05/10 16:09:03 rsh Exp $";   /* RCS ID for ident(1) */
 
 /* Global variables */
 struct host_entry *rrlist = NULL;	/* Round-robin linked list "the list" */
@@ -157,8 +66,14 @@ int source_port = DEFAULT_SOURCE_PORT;	/* UDP source port */
 int dest_port = DEFAULT_DEST_PORT;	/* UDP destination port */
 unsigned lifetime = DEFAULT_LIFETIME;	/* Lifetime in seconds */
 int auth_method = DEFAULT_AUTH_METHOD;	/* Authentication method */
-unsigned end_wait = DEFAULT_END_WAIT;	/* Time to wait after all done */
+unsigned end_wait = DEFAULT_END_WAIT;	/* Time to wait after all done in ms */
 unsigned pattern_fuzz = DEFAULT_PATTERN_FUZZ; /* Pattern matching fuzz in ms */
+int dhgroup = DEFAULT_DH_GROUP;		/* Diffie Hellman Group */
+int idtype = DEFAULT_IDTYPE;		/* IKE Identification type */
+char id_string[MAXLINE];
+int id_data_len;
+uint32_t kx_data[48];		/* Key exchange data.  48 is largest needed */
+int kx_data_len;		/* Key exchange data len in 32bit longwords */
 int verbose=0;
 char vendor_id[MAXLINE];		/* Vendor ID string */
 int vendor_id_flag = 0;			/* Indicates if VID to be used */
@@ -169,12 +84,15 @@ int patterns_loaded = 0;		/* Indicates if backoff patterns loaded */
 int trans_enc;				/* Custom transform encrypt */
 int trans_hash;				/* Custom transform hash */
 int trans_auth;				/* Custom transform auth */
-int trans_group;			/* Custom transform group */
+int trans_group;			/* Custom transform DH group */
 struct timeval last_packet_time;	/* Time last packet was sent */
 struct timeval last_recv_time;		/* Time last packet was received */
 struct isakmp_hdr hdr;			/* ISAKMP Header */
 struct isakmp_sa sa_hdr;		/* SA Header */
 struct isakmp_proposal sa_prop;		/* Proposal payload */
+struct isakmp_kx kx;			/* Key exchange for agg. mode */
+struct isakmp_nonce nonce;		/* Nonce for agg. mode */
+struct isakmp_id id;			/* ID for agg. mode */
 struct transform
 {
    struct isakmp_transform trans_hdr;
@@ -183,9 +101,9 @@ struct transform
 };
 struct transform trans[8];		/* Transform payload */
 struct isakmp_vid vid_hdr;		/* Vendor ID header */
-unsigned char vid_md5[16];		/* Vendor ID data - md5 digest */
+md5_byte_t vid_md5[16];			/* Vendor ID data - md5 digest */
 
-char *auth_methods[] = { /* Authentication methods from RFC 2409 Appendix A */
+const char *auth_methods[] = { /* Authentication methods from RFC 2409 Appendix A */
    "UNSPECIFIED",		/* 0 */
    "pre-shared key",		/* 1 */
    "DSS signatures",		/* 2 */
@@ -194,7 +112,7 @@ char *auth_methods[] = { /* Authentication methods from RFC 2409 Appendix A */
    "Revised encryption with RSA" /* 5 */
 };
 
-char *notification_msg[] = { /* Notify Message Types from RFC 2408 3.14.1 */
+const char *notification_msg[] = { /* Notify Message Types from RFC 2408 3.14.1 */
    "UNSPECIFIED",                    /* 0 */
    "INVALID-PAYLOAD-TYPE",           /* 1 */
    "DOI-NOT-SUPPORTED",              /* 2 */
@@ -228,7 +146,7 @@ char *notification_msg[] = { /* Notify Message Types from RFC 2408 3.14.1 */
    "UNEQUAL-PAYLOAD-LENGTHS"         /* 30 */
 };
 
-char *payload_name[] = {     /* Payload types from RFC 2408 3.1 */
+const char *payload_name[] = {     /* Payload types from RFC 2408 3.1 */
    "NONE",                           /* 0 */
    "Security Association",           /* 1 */
    "Proposal",                       /* 2 */
@@ -265,9 +183,13 @@ main(int argc, char *argv[]) {
       {"trans", required_argument, 0, 'a'},
       {"showbackoff", optional_argument, 0, 'o'},
       {"fuzz", required_argument, 0, 'u'},
+      {"id", required_argument, 0, 'n'},
+      {"idtype", required_argument, 0, 'y'},
+      {"dhgroup", required_argument, 0, 'g'},
+      {"patterns", required_argument, 0, 'p'},
       {0, 0, 0, 0}
    };
-   char *short_options = "f:hs:d:r:t:i:b:w:vl:m:Ve:a:o::u:";
+   const char *short_options = "f:hs:d:r:t:i:b:w:vl:m:Ve:a:o::u:n:y:g:p:";
    int arg;
    char arg_str[MAXLINE];	/* Args as string for syslog */
    int options_index=0;
@@ -282,69 +204,82 @@ main(int argc, char *argv[]) {
    struct host_entry *temp_cursor;
    struct hostent *hp;
    struct timeval diff;		/* Difference between two timevals */
-   unsigned long loop_timediff;
-   unsigned long host_timediff;
-   unsigned long end_timediff=0;
+   unsigned long loop_timediff;	/* Time since last packet sent in ms */
+   unsigned long host_timediff; /* Time since last packet sent to this host */
+   unsigned long end_timediff=0; /* Time since last packet received in ms */
+   int arg_str_space;		/* Used to avoid buffer overruns when copying */
+   char patfile[MAXLINE];	/* IKE Backoff pattern file name */
 /*
- *	Open syslog channel and log arguments if required
+ *	Open syslog channel and log arguments if required.
+ *	We must be careful here to avoid overflowing the arg_str buffer
+ *	which could result in a buffer overflow vulnerability.
  */
 #ifdef SYSLOG
-   openlog(basename(argv[0]), LOG_PID, SYSLOG_FACILITY);
+   openlog("ike-scan", LOG_PID, SYSLOG_FACILITY);
    arg_str[0] = '\0';
+   arg_str_space = MAXLINE;	/* Amount of space in the arg_str buffer */
    for (arg=0; arg<argc; arg++) {
-      strcat(arg_str, argv[arg]);
-      if (arg < (argc-1)) {
-         strcat(arg_str, " ");
+      arg_str_space -= strlen(argv[arg]);
+      if (arg_str_space > 0) {
+         strncat(arg_str, argv[arg], arg_str_space);
+         if (arg < (argc-1)) {
+            strcat(arg_str, " ");
+            arg_str_space--;
+         }
       }
    }
    info_syslog("Starting: %s", arg_str);
 #endif
+/*
+ *	Initialise IKE pattern file name to the empty string.
+ */
+   patfile[0] = '\0';
 /*
  *	Process options and arguments.
  */
    while ((arg=getopt_long_only(argc, argv, short_options, long_options, &options_index)) != -1) {
       switch (arg) {
-         MD5_CTX context;
+         md5_state_t context;
          int i;
-         case 'f':
+         case 'f':	/* --file */
             strncpy(filename, optarg, MAXLINE);
             filename_flag=1;
             break;
-         case 'h':
+         case 'h':	/* --help */
             usage();
             break;
-         case 's':
+         case 's':	/* --sport */
             source_port=atoi(optarg);
             break;
-         case 'd':
+         case 'd':	/* --dport */
             dest_port=atoi(optarg);
             break;
-         case 'r':
+         case 'r':	/* --retry */
             retry=atoi(optarg);
             break;
-         case 't':
+         case 't':	/* --timeout */
             timeout=atoi(optarg);
             break;
-         case 'i':
+         case 'i':	/* --interval */
             interval=atoi(optarg);
             break;
-         case 'b':
+         case 'b':	/* --backoff */
             backoff=atof(optarg);
             break;
-         case 'w':
+         case 'w':	/* --selectwait */
             select_timeout=atoi(optarg);
             break;
-         case 'v':
+         case 'v':	/* --verbose */
             verbose++;
             break;
-         case 'l':
+         case 'l':	/* --lifetime */
             lifetime=atoi(optarg);
             break;
-         case 'm':
+         case 'm':	/* --auth */
             auth_method=atoi(optarg);
             break;
-         case 'V':
-            fprintf(stderr, "%s\n\n", VERSION);
+         case 'V':	/* --version */
+            fprintf(stderr, "%s\n\n", PACKAGE_STRING);
             fprintf(stderr, "Copyright (C) 2003 Roy Hills, NTA Monitor Ltd.\n");
             fprintf(stderr, "ike-scan comes with NO WARRANTY to the extent permitted by law.\n");
             fprintf(stderr, "You may redistribute copies of ike-scan under the terms of the GNU\n");
@@ -355,23 +290,23 @@ main(int argc, char *argv[]) {
             fprintf(stderr, "%s\n", rcsid);
             exit(0);
             break;
-         case 'e':
+         case 'e':	/* --vendor */
             strncpy(vendor_id, optarg, MAXLINE);
             vendor_id_flag=1;
-            MD5Init(&context);
-            MD5Update(&context, vendor_id, strlen(vendor_id));
-            MD5Final(&vid_md5,&context);
+            md5_init(&context);
+            md5_append(&context, (const md5_byte_t *)vendor_id, strlen(vendor_id));
+            md5_finish(&context, vid_md5);
             printf("vid_md5: ");
             for (i=0; i<16; i++)
                printf("%.2x",vid_md5[i]);
             printf("\n");
             break;
-         case 'a':
+         case 'a':	/* --trans */
             strncpy(trans_str, optarg, MAXLINE);
             trans_flag=1;
             sscanf(trans_str, "%d,%d,%d,%d", &trans_enc, &trans_hash, &trans_auth, &trans_group);
             break;
-         case 'o':
+         case 'o':	/* --showbackoff */
             showbackoff_flag=1;
             if (optarg == NULL) {
                end_wait=1000 * DEFAULT_END_WAIT;
@@ -379,10 +314,22 @@ main(int argc, char *argv[]) {
                end_wait=1000 * atoi(optarg);
             }
             break;
-         case 'u':
+         case 'u':	/* --fuzz */
             pattern_fuzz=atoi(optarg);
             break;
-         default:
+         case 'n':	/* --id */
+            /* Not implemented yet */
+            break;
+         case 'y':	/* --idtype */
+            /* Not implemented yet */
+            break;
+         case 'g':	/* --dhgroup */
+            /* Not implemented yet */
+            break;
+         case 'p':	/* --patterns */
+            strncpy(patfile, optarg, MAXLINE);
+            break;
+         default:	/* Unknown option */
             usage();
             break;
       }
@@ -434,10 +381,29 @@ main(int argc, char *argv[]) {
    if (showbackoff_flag) {
       FILE *fp;
       char line[MAXLINE];
-      char patfile[MAXLINE];
       int line_no;
+#ifdef __CYGWIN__
+      char fnbuf[MAXLINE];
+      int fnbuf_siz;
+      int i;
+#endif
+
+      if (patfile[0] == '\0') {	/* If patterns file not specified */
+#ifdef __CYGWIN__
+         if ((fnbuf_siz=GetModuleFileName(GetModuleHandle(0), fnbuf, MAXLINE)) == 0) {
+            err_msg("Call to GetModuleFileName failed");
+         }
+         for (i=fnbuf_siz-1; i>=0 && fnbuf[i] != '/' && fnbuf[i] != '\\'; i--)
+            ;
+         if (i >= 0) {
+            fnbuf[i] = '\0';
+         }
+         sprintf(patfile, "%s\\%s", fnbuf, PATTERNS_FILE);
+#else
+         sprintf(patfile, "%s/%s", IKEDATADIR, PATTERNS_FILE);
+#endif
+      }
 
-      sprintf(patfile, "%s/%s", DATADIR, PATTERNS_FILE);
       if ((fp = fopen(patfile, "r")) == NULL) {
          warn_msg("WARNING: Cannot open IKE backoff patterns file.  ike-scan will still display");
          warn_msg("the backoff patterns, but it will not be able to identify the fingerprints.");
@@ -491,7 +457,7 @@ main(int argc, char *argv[]) {
 /*
  *	Display initial message.
  */
-   printf("Starting %s with %d hosts (http://www.nta-monitor.com/ike-scan/)\n", VERSION, num_hosts);
+   printf("Starting %s with %u hosts (http://www.nta-monitor.com/ike-scan/)\n", PACKAGE_STRING, num_hosts);
 /*
  *	Display the lists if verbose setting is 3 or more.
  */
@@ -505,9 +471,11 @@ main(int argc, char *argv[]) {
  *
  *	The loop exits when all hosts have either responded or timed out
  *	and, if showbackoff_flag is set, at least end_wait ms have elapsed
- *	since the last packet was received.
+ *	since the last packet was received and we have received at least one
+ *	transform response.
  */
-   while (live_count || (showbackoff_flag && end_timediff < end_wait)) {
+   while (live_count ||
+          (showbackoff_flag && transform_responders && (end_timediff < end_wait))) {
 /*
  *	Obtain current time and calculate deltas since last packet and
  *	last packet to this host.
@@ -540,7 +508,7 @@ main(int argc, char *argv[]) {
  */
             if (cursor->num_sent >= retry) {
                if (verbose)
-                  warn_msg("---\tRemoving host entry %d (%s) - Timeout", cursor->n, inet_ntoa(cursor->addr));
+                  warn_msg("---\tRemoving host entry %u (%s) - Timeout", cursor->n, inet_ntoa(cursor->addr));
                remove_host(cursor);
             } else {	/* Retry limit not reached for this host */
                if (cursor->num_sent) {
@@ -565,36 +533,40 @@ main(int argc, char *argv[]) {
  */
             add_recv_time(temp_cursor);
             if (verbose > 1)
-               warn_msg("---\tReceived packet #%d from %s",temp_cursor->num_recv ,inet_ntoa(sa_peer.sin_addr));
+               warn_msg("---\tReceived packet #%u from %s",temp_cursor->num_recv ,inet_ntoa(sa_peer.sin_addr));
             if (temp_cursor->live) {
                display_packet(n, packet_in, temp_cursor, &(sa_peer.sin_addr));
                if (verbose)
-                  warn_msg("---\tRemoving host entry %d (%s) - Received %d bytes", temp_cursor->n, inet_ntoa(sa_peer.sin_addr), n);
+                  warn_msg("---\tRemoving host entry %u (%s) - Received %d bytes", temp_cursor->n, inet_ntoa(sa_peer.sin_addr), n);
                remove_host(temp_cursor);
             }
          } else {
             struct isakmp_hdr hdr_in;
 /*
- *	The received cookie doesn't match any entry in the list
- *	so just issue a message to that effect and ignore the packet.
+ *	The received cookie doesn't match any entry in the list.
+ *	Issue a message to that effect if verbose is on and ignore the packet.
  */
             if (verbose && n >= sizeof(hdr_in)) {
                memcpy(&hdr_in, packet_in, sizeof(hdr_in));
-               warn_msg("---\tIgnoring %d bytes from %s with unknown cookie %0x%0x", n, inet_ntoa(sa_peer.sin_addr),hdr_in.isa_icookie[0], hdr_in.isa_icookie[1]);
+               warn_msg("---\tIgnoring %d bytes from %s with unknown cookie %.8x%.8x", n, inet_ntoa(sa_peer.sin_addr), (uint32_t) htonl(hdr_in.isa_icookie[0]), (uint32_t) htonl(hdr_in.isa_icookie[1]));
             }
          }
       } /* End If */
    } /* End While */
 /*
- *	Display the backoff times if --showbackoff option was specified.
+ *	Display the backoff times if --showbackoff option was specified
+ *	and we have at least one system returning a handshake.
  */
-   if (showbackoff_flag)
+   printf("\n");	/* Ensure we have a blank line */
+   if (showbackoff_flag && transform_responders) {
       dump_times();
+   }
 
    close(sockfd);
 #ifdef SYSLOG
-   info_syslog("Ending: %d hosts scanned. %d returned transform; %d returned notify", num_hosts, transform_responders, notify_responders);
+   info_syslog("Ending: %u hosts scanned. %u returned handshake; %u returned notify", num_hosts, transform_responders, notify_responders);
 #endif
+   printf("Ending %s: %u hosts scanned.  %u returned handshake; %u returned notify\n", PACKAGE_STRING, num_hosts, transform_responders, notify_responders);
    return(0);
 }
 
@@ -607,8 +579,8 @@ add_host(char *name) {
    struct host_entry *he;
    char str[MAXLINE];
    struct timeval now;
-   MD5_CTX context;
-   unsigned char cookie_md5[16];	/* Cookie data - md5 digest */
+   md5_state_t context;
+   md5_byte_t cookie_md5[16];	/* Cookie data - md5 digest */
 
    if ((hp = gethostbyname(name)) == NULL)
       err_sys("gethostbyname");
@@ -631,10 +603,10 @@ add_host(char *name) {
    he->last_send_time.tv_sec=0;
    he->last_send_time.tv_usec=0;
    he->recv_times = NULL;
-   sprintf(str, "%lu %lu %d %s", now.tv_sec, now.tv_usec, num_hosts, inet_ntoa(he->addr));
-   MD5Init(&context);
-   MD5Update(&context, str, strlen(str));
-   MD5Final(&cookie_md5,&context);
+   sprintf(str, "%lu %lu %u %s", now.tv_sec, now.tv_usec, num_hosts, inet_ntoa(he->addr));
+   md5_init(&context);
+   md5_append(&context, (const md5_byte_t *)str, strlen(str));
+   md5_finish(&context, cookie_md5);
    memcpy(he->icookie, cookie_md5, sizeof(he->icookie));
 
    if (rrlist) {	/* List is not empty so add entry */
@@ -745,6 +717,7 @@ display_packet(int n, char *packet_in, struct host_entry *he, struct in_addr *re
    int msg_type;                /* Notification message type */
    char msg_in[MAXLINE];        /* Notification message */
    char ip_str[MAXLINE];	/* IP address(es) to display at start */
+   char xchg_type[MAXLINE];	/* Exchange type string */
    char *cp;
 /*
  *	Write the IP addresses to the output string.
@@ -756,7 +729,7 @@ display_packet(int n, char *packet_in, struct host_entry *he, struct in_addr *re
    *cp = '\0';
 /*
  *	Check that the received packet is at least as big as the ISAKMP
- *	header.
+ *	header before we try to copy it into an ISAKMP header struct.
  */
    if (n < sizeof(hdr_in)) {
       printf("%sShort packet returned (len < ISAKMP header length)\n", ip_str);
@@ -768,12 +741,13 @@ display_packet(int n, char *packet_in, struct host_entry *he, struct in_addr *re
    memcpy(&hdr_in, packet_in, sizeof(hdr_in));
 /*
  *	Check that the initiator cookie in the packet matches what's in the
- *	host entry.
+ *	host entry.  This check should never fail because we wouldn't get here
+ *	unless we'd already matched the cookie.
  */
    if (hdr_in.isa_icookie[0] != he->icookie[0] || hdr_in.isa_icookie[1] != he->icookie[1]) {
       printf("%sReturned icookie doesn't match (received %.8x%.8x; expected %.8x%.8x)\n",
-         ip_str, htonl(hdr_in.isa_icookie[0]), htonl(hdr_in.isa_icookie[1]),
-         htonl(he->icookie[0]), htonl(he->icookie[1]));
+         ip_str, (uint32_t) htonl(hdr_in.isa_icookie[0]), (uint32_t) htonl(hdr_in.isa_icookie[1]),
+         (uint32_t) htonl(he->icookie[0]), (uint32_t) htonl(he->icookie[1]));
       return;
    }
 
@@ -788,8 +762,15 @@ display_packet(int n, char *packet_in, struct host_entry *he, struct in_addr *re
          packet_in += sizeof(sa_hdr_in);
          memcpy(&sa_prop_in, packet_in, sizeof(sa_prop_in));
          packet_in += sizeof(sa_prop_in);
+         if (hdr_in.isa_xchg == ISAKMP_XCHG_IDPROT) {	/* Main mode */
+            strcpy(xchg_type, "Main Mode");
+         } else if (hdr_in.isa_xchg == ISAKMP_XCHG_AGGR) {	/* Aggressive */
+            strcpy(xchg_type, "Aggressive Mode");
+         } else {
+            sprintf(xchg_type, "UNKNOWN Mode (%u)", hdr_in.isa_xchg);
+         }
          decode_transform(packet_in, n, sa_prop_in.isap_notrans);
-         printf("%sIKE Handshake returned (%d transforms)\n", ip_str, sa_prop_in.isap_notrans);
+         printf("%sIKE %s Handshake returned (%d transforms)\n", ip_str, xchg_type, sa_prop_in.isap_notrans);
       } else {
          printf("%sIKE Handshake returned (%d byte packet too short to decode)\n", ip_str, n);
       }
@@ -804,7 +785,7 @@ display_packet(int n, char *packet_in, struct host_entry *he, struct in_addr *re
          msg_type = ntohs(notification_in.isan_type);
          if (msg_type < 31) {                /* RFC Defined message types */
             printf("%sNotify message %d (%s)\n", ip_str, msg_type, notification_msg[msg_type]);
-         } else if (msg_type == 9101) {      /* Firewall-1 4.x message */
+         } else if (msg_type == 9101) {      /* Firewall-1 4.x/NG Base msg */
             char *p;
             msg_len = ntohs(notification_in.isan_length) - sizeof(notification_in);
             packet_in += sizeof(notification_in);
@@ -818,7 +799,7 @@ display_packet(int n, char *packet_in, struct host_entry *he, struct in_addr *re
                if (!isprint(*p))
                   *p='.';
             }
-            printf("%sNotify message %d (%s)\n", ip_str, msg_type, msg_in);
+            printf("%sNotify message %d [Checkpoint Firewall-1 4.x or NG Base] (%s)\n", ip_str, msg_type, msg_in);
          } else {                            /* Unknown message type */
             printf("%sNotify message %d (UNKNOWN MESSAGE TYPE)\n", ip_str, msg_type);
          }
@@ -827,7 +808,7 @@ display_packet(int n, char *packet_in, struct host_entry *he, struct in_addr *re
       }
    } else {
 /*
- *	Some other payload that we don't understand.  Display the payload
+ *	Some other payload type that we don't understand.  Display the payload
  *	number, and also the payload name if defined.
  */
       if (hdr_in.isa_np <= MAX_PAYLOAD) {
@@ -858,7 +839,7 @@ send_packet(int s, struct host_entry *he) {
    struct sockaddr_in sa_peer;
    char buf[MAXUDP];
    int buflen;
-   int sa_peer_len;
+   NET_SIZE_T sa_peer_len;
    char *cp;
 /*
  *	Set up the sockaddr_in structure for the host.
@@ -913,7 +894,7 @@ send_packet(int s, struct host_entry *he) {
  *	Send the packet.
  */
    if (verbose > 1)
-      warn_msg("---\tSending packet #%d to host entry %d (%s) tmo %d", he->num_sent, he->n, inet_ntoa(he->addr), he->timeout);
+      warn_msg("---\tSending packet #%u to host entry %u (%s) tmo %d", he->num_sent, he->n, inet_ntoa(he->addr), he->timeout);
    if ((sendto(s, buf, buflen, 0, (struct sockaddr *) &sa_peer, sa_peer_len)) < 0) {
       err_sys("sendto");
    }
@@ -929,7 +910,7 @@ recvfrom_wto(int s, char *buf, int len, struct sockaddr *saddr, int tmo) {
    fd_set readset;
    struct timeval to;
    int n;
-   int saddr_len;
+   NET_SIZE_T saddr_len;
 
    FD_ZERO(&readset);
    FD_SET(s, &readset);
@@ -1237,10 +1218,10 @@ dump_list(void) {
    printf("Host List:\n\n");
    printf("Entry\tIP Address\tCookie\n");
    do {
-      printf("%d\t%s\t%0x%0x\n", p->n, inet_ntoa(p->addr), htonl(p->icookie[0]), htonl(p->icookie[1]));
+      printf("%u\t%s\t%.8x%.8x\n", p->n, inet_ntoa(p->addr), (uint32_t) htonl(p->icookie[0]), (uint32_t) htonl(p->icookie[1]));
       p = p->next;
    } while (p != rrlist);
-   printf("\nTotal of %d host entries.\n\n", num_hosts);
+   printf("\nTotal of %u host entries.\n\n", num_hosts);
 }
 
 /*
@@ -1285,7 +1266,7 @@ dump_times(void) {
 
    p = rrlist;
 
-   printf("\nIKE Backoff Patterns:\n");
+   printf("IKE Backoff Patterns:\n");
    printf("\nIP Address\tNo.\tRecv time\t\tDelta Time\n");
    do {
       if (p->recv_times != NULL && p->num_recv > 1) {
@@ -1296,7 +1277,7 @@ dump_times(void) {
          while (te != NULL) {
             if (i > 1)
                timeval_diff(&(te->time), &prev_time, &diff);
-            printf("%s\t%d\t\%ld.%.6ld\t%ld.%.6ld\n", inet_ntoa(p->addr), i, (long)te->time.tv_sec, (long)te->time.tv_usec, (long)diff.tv_sec, (long)diff.tv_usec);
+            printf("%s\t%d\t%ld.%.6ld\t%ld.%.6ld\n", inet_ntoa(p->addr), i, (long)te->time.tv_sec, (long)te->time.tv_usec, (long)diff.tv_sec, (long)diff.tv_usec);
             prev_time = te->time;
             te = te->next;
             i++;
@@ -1538,20 +1519,20 @@ usage(void) {
    fprintf(stderr, "\n--file=<fn> or -f <fn>\tRead hostnames or addresses from the specified file\n");
    fprintf(stderr, "\t\t\tinstead of from the command line. One name or IP\n");
    fprintf(stderr, "\t\t\taddress per line.  Use \"-\" for standard input.\n");
-   fprintf(stderr, "\n--sport=<p> or -s p\tSet UDP source port to <p>, default=%d, 0=random.\n", DEFAULT_SOURCE_PORT);
+   fprintf(stderr, "\n--sport=<p> or -s <p>\tSet UDP source port to <p>, default=%d, 0=random.\n", DEFAULT_SOURCE_PORT);
    fprintf(stderr, "\t\t\tSome IKE implementations require the client to use\n");
    fprintf(stderr, "\t\t\tUDP source port 500 and will not talk to other ports.\n");
    fprintf(stderr, "\t\t\tNote that superuser privileges are normally required\n");
    fprintf(stderr, "\t\t\tto use non-zero source ports below 1024.  Also only\n");
    fprintf(stderr, "\t\t\tone process on a system may bind to a given source port\n");
    fprintf(stderr, "\t\t\tat any one time.\n");
-   fprintf(stderr, "\n--dport=<p> or -d p\tSet UDP destination port to <p>, default=%d.\n", DEFAULT_DEST_PORT);
+   fprintf(stderr, "\n--dport=<p> or -d <p>\tSet UDP destination port to <p>, default=%d.\n", DEFAULT_DEST_PORT);
    fprintf(stderr, "\t\t\tUDP port 500 is the assigned port number for ISAKMP\n");
    fprintf(stderr, "\t\t\tand this is the port used by most if not all IKE\n");
    fprintf(stderr, "\t\t\timplementations.\n");
-   fprintf(stderr, "\n--retry=<n> or -r n\tSet total number of attempts per host to <n>,\n");
+   fprintf(stderr, "\n--retry=<n> or -r <n>\tSet total number of attempts per host to <n>,\n");
    fprintf(stderr, "\t\t\tdefault=%d.\n", DEFAULT_RETRY);
-   fprintf(stderr, "\n--timeout=<n> or -t n\tSet initial per host timeout to <n> ms, default=%d.\n", DEFAULT_TIMEOUT);
+   fprintf(stderr, "\n--timeout=<n> or -t <n>\tSet initial per host timeout to <n> ms, default=%d.\n", DEFAULT_TIMEOUT);
    fprintf(stderr, "\t\t\tThis timeout is for the first packet sent to each host.\n");
    fprintf(stderr, "\t\t\tsubsequent timeouts are multiplied by the backoff\n");
    fprintf(stderr, "\t\t\tfactor which is set with --backoff.\n");
@@ -1620,8 +1601,18 @@ usage(void) {
    fprintf(stderr, "\t\t\tthe backoff patterns file.  Larger values allow for\n");
    fprintf(stderr, "\t\t\thigher variance but also increase the risk of\n");
    fprintf(stderr, "\t\t\tfalse positive identifications.\n");
+#ifdef __CYGWIN__
+   fprintf(stderr, "\n--patterns=<f> or -p <f> Use IKE patterns file <f>,\n");
+   fprintf(stderr, "\t\t\tdefault=%s in ike-scan.exe dir.\n", PATTERNS_FILE);
+#else
+   fprintf(stderr, "\n--patterns=<f> or -p <f> Use IKE patterns file <f>,\n");
+   fprintf(stderr, "\t\t\tdefault=%s/%s.\n", IKEDATADIR, PATTERNS_FILE);
+#endif
+   fprintf(stderr, "\t\t\tThis specifies the name of the file containing\n");
+   fprintf(stderr, "\t\t\tIKE backoff patterns.  This file is only used when\n");
+   fprintf(stderr, "\t\t\t--showbackoff is specified.\n");
    fprintf(stderr, "\n");
-   fprintf(stderr, "Report bugs or send suggestions to ike-scan@nta-monitor.com\n");
+   fprintf(stderr, "Report bugs or send suggestions to %s\n", PACKAGE_BUGREPORT);
    fprintf(stderr, "See the ike-scan homepage at http://www.nta-monitor.com/ike-scan/\n");
    exit(1);
 }
