@@ -21,7 +21,7 @@
  * You are encouraged to send comments, improvements or suggestions to
  * me at ike-scan@nta-monitor.com.
  *
- * $Id: ike-scan.h,v 1.27 2003/01/15 13:59:29 rsh Exp $
+ * $Id: ike-scan.h,v 1.36 2003/05/09 15:49:17 rsh Exp $
  *
  * ike-scan.h -- Header file for IKE Scanner (ike-scan)
  *
@@ -30,12 +30,75 @@
  */
 
 /* Includes */
-#include <sys/types.h>  /* FreeBSD needs explicit include for sys/types.h */
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
+
+#ifdef STDC_HEADERS
+#include <stdio.h>
+#include <string.h>
+#include <stdlib.h>
+#include <ctype.h>
 #include <stdarg.h>
+#include <errno.h>
+#include <math.h>
+#else
+#error This program requires the ANSI C Headers
+#endif
+
+#include <sys/types.h>  /* FreeBSD needs explicit include for sys/types.h */
+
+#ifdef __CYGWIN__
+#include <windows.h>	/* Include windows.h if compiling under Cygwin */
+#endif
+
+#ifdef HAVE_UNISTD_H
+#include <unistd.h>
+#endif
+
+#ifdef HAVE_NETDB_H
+#include <netdb.h>
+#endif
+
+#ifdef HAVE_SYSLOG_H
+#include <syslog.h>
+#endif
+
+#ifdef HAVE_GETOPT_H
+#include <getopt.h>
+#else
+/* Include getopt.h for the sake of getopt_long.
+   We don't need the declaration of getopt, and it could conflict
+   with something from a system header file, so effectively nullify that.  */
+#define getopt getopt_loser
+#include "getopt.h"
+#undef getopt
+#endif
+
+#ifdef HAVE_NETINET_IN_H
 #include <netinet/in.h>
-#include <sys/time.h>
+#endif
+
+#ifdef TIME_WITH_SYS_TIME
+# include <sys/time.h>
+# include <time.h>
+#else
+# ifdef HAVE_SYS_TIME_H
+#  include <sys/time.h>
+# else
+#  include <time.h>
+# endif
+#endif
+
+#ifdef HAVE_SYS_SOCKET_H
 #include <sys/socket.h>	/* For struct sockaddr */
+#endif
+
+#ifdef HAVE_ARPA_INET_H
 #include <arpa/inet.h>
+#endif
+
+#include "md5.h"
 #include "isakmp.h"
 
 /* Defines */
@@ -53,6 +116,8 @@
 #define DEFAULT_AUTH_METHOD 1		/* Default authentication method */
 #define DEFAULT_END_WAIT 60		/* Default time to wait at end in sec */
 #define DEFAULT_PATTERN_FUZZ 100	/* Default pattern fuzz in ms */
+#define DEFAULT_DH_GROUP 2		/* Default DH group for agg. mode */
+#define DEFAULT_IDTYPE ID_USER_FQDN	/* Default ID Type for agg. mode */
 #define SYSLOG 1			/* Use syslog if defined */
 #define SYSLOG_FACILITY LOG_USER	/* Syslog facility to use */
 #define PATTERNS_FILE "ike-backoff-patterns" /* Backoff patterns filename */
@@ -61,7 +126,7 @@
 struct host_entry {
    struct host_entry *prev;	/* Previous pointer */
    struct host_entry *next;	/* Next pointer */
-   int n;			/* Ordinal number for this entry */
+   unsigned n;			/* Ordinal number for this entry */
    struct in_addr addr;		/* Host IP address */
    u_char live;			/* Set when awaiting response */
    struct timeval last_send_time; /* Time when last packet sent to this addr */
